# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Tag definitions are created independent of objects tagged with tags. Primary key of this table is ``oid``.
class TagDefinition(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TagDefinition()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTagDefinition(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # TagDefinition
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of the tag definition within the database.
    # TagDefinition
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TagDefinition
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TagDefinition
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TagDefinition
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Timestamp when the attribute was last modified (or first created).
    # TagDefinition
    def Modified(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Tag name (or URI in general). This is must be globally unique and represents the semantic identifier of the tag.
    # TagDefinition
    def Tag(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # CBOR-serialized, object-valued extra data stored along with this tag definition.
    # TagDefinition
    def Extra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TagDefinition
    def ExtraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TagDefinition
    def ExtraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TagDefinition
    def ExtraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0


def TagDefinitionStart(builder):
    builder.StartObject(4)


def Start(builder):
    TagDefinitionStart(builder)


def TagDefinitionAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    TagDefinitionAddOid(builder, oid)


def TagDefinitionStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return TagDefinitionStartOidVector(builder, numElems)


def TagDefinitionAddModified(builder, modified):
    builder.PrependUint64Slot(1, modified, 0)


def AddModified(builder, modified):
    TagDefinitionAddModified(builder, modified)


def TagDefinitionAddTag(builder, tag):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(tag), 0)


def AddTag(builder, tag):
    TagDefinitionAddTag(builder, tag)


def TagDefinitionAddExtra(builder, extra):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(extra), 0)


def AddExtra(builder, extra):
    TagDefinitionAddExtra(builder, extra)


def TagDefinitionStartExtraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartExtraVector(builder, numElems: int) -> int:
    return TagDefinitionStartExtraVector(builder, numElems)


def TagDefinitionEnd(builder):
    return builder.EndObject()


def End(builder):
    return TagDefinitionEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# WAMP permission (authorization of performing a WAMP action on a WAMP URI pattern).
class Permission(object):
    __slots__ = ["_tab"]

    @classmethod
    def SizeOf(cls):
        return 9

    # Permission
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # URI or URI pattern to match for the permission to apply.
    # URI check level.
    # Permission
    def UriCheckLevel(self):
        return self._tab.Get(
            flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0)
        )

    # URI matching mode.
    # Permission
    def Match(self):
        return self._tab.Get(
            flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1)
        )

    # Allow/disallow calling procedures on a match.
    # Permission
    def AllowCall(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2)
        )

    # Allow/disallow registering procedures on a match.
    # Permission
    def AllowRegister(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3)
        )

    # Allow/disallow publishing events on a match.
    # Permission
    def AllowPublish(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4)
        )

    # Allow/disallow subscribing topics on a match.
    # Permission
    def AllowSubscribe(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(5)
        )

    # Disclose the caller on a match (of a procedure) when called.
    # Permission
    def DiscloseCaller(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(6)
        )

    # Disclose the publisher on a match (of a topic) when published to.
    # Permission
    def DisclosePublisher(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(7)
        )

    # Cache the permission on a match in the router worker.
    # Permission
    def Cache(self):
        return self._tab.Get(
            flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8)
        )


def CreatePermission(
    builder,
    uriCheckLevel,
    match,
    allowCall,
    allowRegister,
    allowPublish,
    allowSubscribe,
    discloseCaller,
    disclosePublisher,
    cache,
):
    builder.Prep(1, 9)
    builder.PrependBool(cache)
    builder.PrependBool(disclosePublisher)
    builder.PrependBool(discloseCaller)
    builder.PrependBool(allowSubscribe)
    builder.PrependBool(allowPublish)
    builder.PrependBool(allowRegister)
    builder.PrependBool(allowCall)
    builder.PrependInt8(match)
    builder.PrependInt8(uriCheckLevel)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: realmstore

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table store WAMP events dispatched to receivers, under WAMP subscriptions on URIs (or patterns).
class Event(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Event()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsEvent(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Event
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Timestamp when the event was sent to the receiver. Epoch time in ns.
    # Event
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The subscription ID this event is dispatched under.
    # Event
    def Subscription(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The publication ID of the dispatched event.
    # Event
    def Publication(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The WAMP session ID of the receiver.
    # Event
    def Receiver(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Whether the message was retained by the broker on the topic, rather than just published.
    # Event
    def Retained(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Whether this Event was to be acknowledged by the receiver.
    # Event
    def AcknowledgedDelivery(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False


def EventStart(builder):
    builder.StartObject(6)


def Start(builder):
    EventStart(builder)


def EventAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    EventAddTimestamp(builder, timestamp)


def EventAddSubscription(builder, subscription):
    builder.PrependUint64Slot(1, subscription, 0)


def AddSubscription(builder, subscription):
    EventAddSubscription(builder, subscription)


def EventAddPublication(builder, publication):
    builder.PrependUint64Slot(2, publication, 0)


def AddPublication(builder, publication):
    EventAddPublication(builder, publication)


def EventAddReceiver(builder, receiver):
    builder.PrependUint64Slot(3, receiver, 0)


def AddReceiver(builder, receiver):
    EventAddReceiver(builder, receiver)


def EventAddRetained(builder, retained):
    builder.PrependBoolSlot(4, retained, 0)


def AddRetained(builder, retained):
    EventAddRetained(builder, retained)


def EventAddAcknowledgedDelivery(builder, acknowledgedDelivery):
    builder.PrependBoolSlot(5, acknowledgedDelivery, 0)


def AddAcknowledgedDelivery(builder, acknowledgedDelivery):
    EventAddAcknowledgedDelivery(builder, acknowledgedDelivery)


def EventEnd(builder):
    return builder.EndObject()


def End(builder):
    return EventEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR data service APIs published to (part of) catalogs.
class Api(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Api()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsApi(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Api
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The unique ID of the API.
    # Api
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Api
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Api
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Api
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # The ID of the catalog this API is part of (was published to).
    # Api
    def CatalogOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Api
    def CatalogOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Api
    def CatalogOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Api
    def CatalogOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Database transaction time (epoch time in ns).
    # Api
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number when the catalog was created.
    # Api
    def Published(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Api
    def PublishedAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Api
    def PublishedLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Api
    def PublishedIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Multihash of API Flatbuffers schema (required).
    # Api
    def Schema(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Multihash of API meta-data (optional).
    # Api
    def Meta(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Api
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Api
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Api
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Api
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # When signed off-chain and submitted via ``XBRCatalog.publishApiFor``.
    # Api
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Api
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Api
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Api
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0


def ApiStart(builder):
    builder.StartObject(8)


def Start(builder):
    ApiStart(builder)


def ApiAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    ApiAddOid(builder, oid)


def ApiStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return ApiStartOidVector(builder, numElems)


def ApiAddCatalogOid(builder, catalogOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(catalogOid), 0)


def AddCatalogOid(builder, catalogOid):
    ApiAddCatalogOid(builder, catalogOid)


def ApiStartCatalogOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCatalogOidVector(builder, numElems: int) -> int:
    return ApiStartCatalogOidVector(builder, numElems)


def ApiAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(2, timestamp, 0)


def AddTimestamp(builder, timestamp):
    ApiAddTimestamp(builder, timestamp)


def ApiAddPublished(builder, published):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(published), 0)


def AddPublished(builder, published):
    ApiAddPublished(builder, published)


def ApiStartPublishedVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPublishedVector(builder, numElems: int) -> int:
    return ApiStartPublishedVector(builder, numElems)


def ApiAddSchema(builder, schema):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(schema), 0)


def AddSchema(builder, schema):
    ApiAddSchema(builder, schema)


def ApiAddMeta(builder, meta):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(meta), 0)


def AddMeta(builder, meta):
    ApiAddMeta(builder, meta)


def ApiAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    ApiAddTid(builder, tid)


def ApiStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return ApiStartTidVector(builder, numElems)


def ApiAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    ApiAddSignature(builder, signature)


def ApiStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return ApiStartSignatureVector(builder, numElems)


def ApiEnd(builder):
    return builder.EndObject()


def End(builder):
    return ApiEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import pytest
import txaio
import zlmdb
from zlmdb import flatbuffers

txaio.use_twisted()  # noqa

from txaio import time_ns

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr.consent import Consent

zlmdb.TABLES_BY_UUID = {}


def fill_consent(consent):
    consent.market_oid = uuid.uuid4()
    consent.member = os.urandom(20)
    consent.delegate = os.urandom(20)
    consent.delegate_type = random.randint(1, 3)
    consent.catalog_oid = uuid.uuid4()
    consent.timestamp = np.datetime64(time_ns(), "ns")
    consent.updated = random.randint(1, 2**256 - 1)
    consent.consent = True
    consent.synced = True
    consent.service_prefix = _gen_ipfs_hash()
    consent.tid = os.urandom(32)
    consent.signature = os.urandom(65)


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def consent():
    _consent = Consent()
    fill_consent(_consent)
    return _consent


def test_consent_roundtrip(consent, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = consent.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 392

    # create python object from bytes (flatbuffes)
    _consent = Consent.cast(data)

    assert _consent.market_oid == consent.market_oid
    assert _consent.member == consent.member
    assert _consent.delegate == consent.delegate
    assert _consent.delegate_type == consent.delegate_type
    assert _consent.catalog_oid == consent.catalog_oid
    assert _consent.timestamp == consent.timestamp
    assert _consent.updated == consent.updated
    assert _consent.consent == consent.consent
    assert _consent.service_prefix == consent.service_prefix
    assert _consent.tid == consent.tid
    assert _consent.signature == consent.signature
    assert _consent.synced == consent.synced


def test_consent_roundtrip_perf(consent, builder):
    obj = consent.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _consent = Consent.cast(data)
        if True:
            assert _consent.market_oid == consent.market_oid
            assert _consent.member == consent.member
            assert _consent.delegate == consent.delegate
            assert _consent.delegate_type == consent.delegate_type
            assert _consent.catalog_oid == consent.catalog_oid
            assert _consent.timestamp == consent.timestamp
            assert _consent.updated == consent.updated
            assert _consent.consent == consent.consent
            assert _consent.service_prefix == consent.service_prefix
            assert _consent.tid == consent.tid
            assert _consent.signature == consent.signature
            assert _consent.synced == consent.synced

            scratch["value"] += _consent.updated

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapBytes32FlatBuffers, MapUuidTimestampBytes32, flatbuffers, table

from cfxdb.gen.xbrmm import UserKey as UserKeyGen


class _UserKeyGen(UserKeyGen.UserKey):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsUserKey(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _UserKeyGen()
        x.Init(buf, n + offset)
        return x

    def PubkeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OwnerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def WalletAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class UserKey:
    """
    User client (public) keys.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # User key - a Ed25519 public key - for authenticating using WAMP-cryptosign.
        # [uint8] (uint256)
        self._pubkey = None

        # Timestamp (epoch time in ns) of initial creation of this record.
        # uint64 (timestamp)
        self._created = None

        # ID of user account this user key is owned by.
        # [uint8] (uuid)
        self._owner = None

        # Wallet address of user account this key is owned by.
        # [uint8] (address)
        self._wallet_address = None

        # Signature which resulted in this key being saved.
        # [uint8] (ethsig)
        self._signature = None

    def marshal(self):
        obj = {
            "pubkey": bytes(self.pubkey),
            "created": int(self.created),
            "owner": self.owner.bytes,
            "wallet_address": self.wallet_address if self.wallet_address else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def pubkey(self) -> bytes:
        """
        User key - a Ed25519 public key - for authenticating using WAMP-cryptosign.
        """
        if self._pubkey is None and self._from_fbs:
            if self._from_fbs.PubkeyLength():
                self._pubkey = self._from_fbs.PubkeyAsBytes()
        return self._pubkey

    @pubkey.setter
    def pubkey(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._pubkey = value

    @property
    def created(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) of initial creation of this record.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def owner(self) -> uuid.UUID:
        """
        ID of user account this user key is owned by.
        """
        if self._owner is None and self._from_fbs:
            if self._from_fbs.OwnerLength():
                _owner = self._from_fbs.OwnerAsBytes()
                self._owner = uuid.UUID(bytes=bytes(_owner))
            else:
                self._owner = uuid.UUID(bytes=b"\x00" * 20)
        return self._owner

    @owner.setter
    def owner(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._owner = value

    @property
    def wallet_address(self) -> bytes:
        """
        Wallet address of the member this key belong to.
        """
        if self._wallet_address is None and self._from_fbs:
            if self._from_fbs.WalletAddressLength():
                self._wallet_address = self._from_fbs.WalletAddressAsBytes()
        return self._wallet_address

    @wallet_address.setter
    def wallet_address(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._wallet_address = value

    @property
    def signature(self) -> bytes:
        """
        market maker transaction signature.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return UserKey(_UserKeyGen.GetRootAsUserKey(buf, 0))

    def build(self, builder):
        owner = self.owner.bytes if self.owner else None
        if owner:
            owner = builder.CreateString(owner)

        pubkey = self.pubkey
        if pubkey:
            pubkey = builder.CreateString(pubkey)

        wallet_address = self.wallet_address
        if wallet_address:
            wallet_address = builder.CreateString(wallet_address)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        UserKeyGen.UserKeyStart(builder)

        if pubkey:
            UserKeyGen.UserKeyAddPubkey(builder, pubkey)

        if self.created:
            UserKeyGen.UserKeyAddCreated(builder, int(self.created))

        if owner:
            UserKeyGen.UserKeyAddOwner(builder, owner)

        if wallet_address:
            UserKeyGen.UserKeyAddWalletAddress(builder, wallet_address)

        if signature:
            UserKeyGen.UserKeyAddSignature(builder, signature)

        final = UserKeyGen.UserKeyEnd(builder)

        return final


@table("62d140f6-c1f4-4f38-bfde-e450f29d0f95", build=UserKey.build, cast=UserKey.cast)
class UserKeys(MapBytes32FlatBuffers):
    """
    Database table for user client keys.
    """


@table("d4476238-c642-4625-8844-780137152ddd")
class IndexUserKeyByMember(MapUuidTimestampBytes32):
    """
    Database (index) table for (member_oid, created) -> userkey mapping.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class WebCluster(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = WebCluster()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsWebCluster(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # WebCluster
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # WebCluster
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Label for this object (not interpreted by CFC).
    # WebCluster
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Description for this object (not interpreted by CFC).
    # WebCluster
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tags on this object.
    # WebCluster
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # WebCluster
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # WebCluster
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Unique user assigned name.
    # WebCluster
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Current status of web cluster.
    # WebCluster
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # IP version, either 4 for 6
    # WebCluster
    def TcpVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # IP listening port
    # WebCluster
    def TcpPort(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # enable TCP port sharing
    # WebCluster
    def TcpShared(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # listen on this interface
    # WebCluster
    def TcpInterface(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # TCP accept backlog queue size
    # WebCluster
    def TcpBacklog(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TLS server private key to use
    # WebCluster
    def TlsKey(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # TLS server certificate to use
    # WebCluster
    def TlsCertificate(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # TLS certificate chain
    # WebCluster
    def TlsChainCertificates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # WebCluster
    def TlsChainCertificatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # WebCluster
    def TlsChainCertificatesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

    # CA certificates to use
    # WebCluster
    def TlsCaCertificates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # WebCluster
    def TlsCaCertificatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # WebCluster
    def TlsCaCertificatesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # DH parameter file
    # WebCluster
    def TlsCiphers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Ciphers list
    # WebCluster
    def TlsDhparam(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # HTTP client inactivity timeout
    # WebCluster
    def HttpClientTimeout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # enable HTTP strict transport security (HSTS)
    # WebCluster
    def HttpHsts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # HSTS maximum age to announce
    # WebCluster
    def HttpHstsMaxAge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # enable Web request access logging
    # WebCluster
    def HttpAccessLog(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # enable tracebacks when running into Web errors
    # WebCluster
    def HttpDisplayTracebacks(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False


def WebClusterStart(builder):
    builder.StartObject(22)


def Start(builder):
    WebClusterStart(builder)


def WebClusterAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    WebClusterAddOid(builder, oid)


def WebClusterAddLabel(builder, label):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)


def AddLabel(builder, label):
    WebClusterAddLabel(builder, label)


def WebClusterAddDescription(builder, description):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)


def AddDescription(builder, description):
    WebClusterAddDescription(builder, description)


def WebClusterAddTags(builder, tags):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)


def AddTags(builder, tags):
    WebClusterAddTags(builder, tags)


def WebClusterStartTagsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTagsVector(builder, numElems: int) -> int:
    return WebClusterStartTagsVector(builder, numElems)


def WebClusterAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    WebClusterAddName(builder, name)


def WebClusterAddStatus(builder, status):
    builder.PrependUint8Slot(5, status, 0)


def AddStatus(builder, status):
    WebClusterAddStatus(builder, status)


def WebClusterAddTcpVersion(builder, tcpVersion):
    builder.PrependInt8Slot(6, tcpVersion, 0)


def AddTcpVersion(builder, tcpVersion):
    WebClusterAddTcpVersion(builder, tcpVersion)


def WebClusterAddTcpPort(builder, tcpPort):
    builder.PrependUint16Slot(7, tcpPort, 0)


def AddTcpPort(builder, tcpPort):
    WebClusterAddTcpPort(builder, tcpPort)


def WebClusterAddTcpShared(builder, tcpShared):
    builder.PrependBoolSlot(8, tcpShared, 0)


def AddTcpShared(builder, tcpShared):
    WebClusterAddTcpShared(builder, tcpShared)


def WebClusterAddTcpInterface(builder, tcpInterface):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(tcpInterface), 0)


def AddTcpInterface(builder, tcpInterface):
    WebClusterAddTcpInterface(builder, tcpInterface)


def WebClusterAddTcpBacklog(builder, tcpBacklog):
    builder.PrependInt32Slot(10, tcpBacklog, 0)


def AddTcpBacklog(builder, tcpBacklog):
    WebClusterAddTcpBacklog(builder, tcpBacklog)


def WebClusterAddTlsKey(builder, tlsKey):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(tlsKey), 0)


def AddTlsKey(builder, tlsKey):
    WebClusterAddTlsKey(builder, tlsKey)


def WebClusterAddTlsCertificate(builder, tlsCertificate):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(tlsCertificate), 0)


def AddTlsCertificate(builder, tlsCertificate):
    WebClusterAddTlsCertificate(builder, tlsCertificate)


def WebClusterAddTlsChainCertificates(builder, tlsChainCertificates):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(tlsChainCertificates), 0)


def AddTlsChainCertificates(builder, tlsChainCertificates):
    WebClusterAddTlsChainCertificates(builder, tlsChainCertificates)


def WebClusterStartTlsChainCertificatesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTlsChainCertificatesVector(builder, numElems: int) -> int:
    return WebClusterStartTlsChainCertificatesVector(builder, numElems)


def WebClusterAddTlsCaCertificates(builder, tlsCaCertificates):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(tlsCaCertificates), 0)


def AddTlsCaCertificates(builder, tlsCaCertificates):
    WebClusterAddTlsCaCertificates(builder, tlsCaCertificates)


def WebClusterStartTlsCaCertificatesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTlsCaCertificatesVector(builder, numElems: int) -> int:
    return WebClusterStartTlsCaCertificatesVector(builder, numElems)


def WebClusterAddTlsCiphers(builder, tlsCiphers):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(tlsCiphers), 0)


def AddTlsCiphers(builder, tlsCiphers):
    WebClusterAddTlsCiphers(builder, tlsCiphers)


def WebClusterAddTlsDhparam(builder, tlsDhparam):
    builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(tlsDhparam), 0)


def AddTlsDhparam(builder, tlsDhparam):
    WebClusterAddTlsDhparam(builder, tlsDhparam)


def WebClusterAddHttpClientTimeout(builder, httpClientTimeout):
    builder.PrependInt32Slot(17, httpClientTimeout, 0)


def AddHttpClientTimeout(builder, httpClientTimeout):
    WebClusterAddHttpClientTimeout(builder, httpClientTimeout)


def WebClusterAddHttpHsts(builder, httpHsts):
    builder.PrependBoolSlot(18, httpHsts, 0)


def AddHttpHsts(builder, httpHsts):
    WebClusterAddHttpHsts(builder, httpHsts)


def WebClusterAddHttpHstsMaxAge(builder, httpHstsMaxAge):
    builder.PrependInt32Slot(19, httpHstsMaxAge, 0)


def AddHttpHstsMaxAge(builder, httpHstsMaxAge):
    WebClusterAddHttpHstsMaxAge(builder, httpHstsMaxAge)


def WebClusterAddHttpAccessLog(builder, httpAccessLog):
    builder.PrependBoolSlot(20, httpAccessLog, 0)


def AddHttpAccessLog(builder, httpAccessLog):
    WebClusterAddHttpAccessLog(builder, httpAccessLog)


def WebClusterAddHttpDisplayTracebacks(builder, httpDisplayTracebacks):
    builder.PrependBoolSlot(21, httpDisplayTracebacks, 0)


def AddHttpDisplayTracebacks(builder, httpDisplayTracebacks):
    WebClusterAddHttpDisplayTracebacks(builder, httpDisplayTracebacks)


def WebClusterEnd(builder):
    return builder.EndObject()


def End(builder):
    return WebClusterEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import cbor2
import numpy as np
from zlmdb import MapUuidFlatBuffers, flatbuffers, table

from cfxdb.gen.xbrnetwork import VerifiedAction as VerifiedActionGen


class _VerifiedActionGen(VerifiedActionGen.VerifiedAction):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsVerifiedAction(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _VerifiedActionGen()
        x.Init(buf, n + offset)
        return x

    def OidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def VerifiedOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def VerifiedDataAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class VerificationType(object):
    NONE = 0
    """
    Unset
    """

    MEMBER_ONBOARD_EMAIL = 1
    """
    Member onboarded, using verification via member primary email.
    """

    MEMBER_LOGIN_EMAIL = 2
    """
    Member created a market, using verification via member primary email.
    """

    CREATE_MARKET_EMAIL = 3
    """
    Member joined a market, using verification via member primary email.
    """


class VerifiedAction(object):
    """
    User actions (such as "on-board new user") yet to be verified. Actions to be verified are
    identified using their "Verification Action ID". A verificatin email is sent to the (primary)
    email address of the user, and to verify, the user must click the verification link contained
    in the email. The verification link contains the verification action ID and code.
    """

    VERIFICATION_TYPE_NONE = 0
    """
    Unset verification action type ("null").
    """

    VERIFICATION_TYPE_ONBOARD_MEMBER = 1
    """
    Verification action type for on-boarding new members via email verification.
    """

    VERIFICATION_TYPE_LOGIN_MEMBER = 2
    """
    Verification action type for login of member (client) via email verification.
    """

    VERIFICATION_TYPE_CREATE_MARKET = 3
    """
    Verification action type for creation of new data market via email verification.
    """

    VERIFICATION_TYPE_JOIN_MARKET = 4
    """
    Verification action type for joining an existing market via email verification.
    """

    VERIFICATION_TYPE_CREATE_CATALOG = 5
    """
    Verification action type for creation of a catalog via email verification.
    """

    VERIFICATION_TYPE_PUBLISH_API = 5
    """
    Verification action type for publishing on API via email verification.
    """

    VERIFICATION_TYPE = list(range(7))
    """
    All valid verification action types.
    """

    VERIFICATION_STATUS_NONE = 0
    """
    Unset verification action status ("null").
    """

    VERIFICATION_STATUS_PENDING = 1
    """
    Verification is still pending.
    """

    VERIFICATION_STATUS_VERIFIED = 2
    """
    Verification has been successfully completed.
    """

    VERIFICATION_STATUS_FAILED = 3
    """
    Verification has failed.
    """

    VERIFICATION_STATUS_EXPIRED = 4
    """
    Action cannot be verified any longer since it has expired.
    """

    VERIFICATION_STATUS = list(range(5))
    """
    All valid verification status codes.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # Globally unique and static ID of action .
        # [uint8] (uuid)
        self._oid = None

        # Timestamp (epoch time in ns) of initial creation of this record.
        # uint64 (timestamp)
        self._created = None

        # Verification type.
        # VerificationType (uint8)
        self._vtype = None

        # Verification type.
        # VerificationStatus (uint8)
        self._vstatus = None

        # Verification code sent.
        # string
        self._vcode = None

        # ID of object of verified action.
        # [uint8] (uuid)
        self._verified_oid = None

        # Action data, serialized in CBOR.
        # [uint8]
        self._verified_data = None

    def marshal(self):
        obj = {
            "oid": self.oid.bytes if self.oid else None,
            "created": int(self.created) if self.created else None,
            "vtype": self.vtype,
            "vstatus": self.vstatus,
            "vcode": self.vcode,
            "verified_oid": self.verified_oid.bytes if self.verified_oid.bytes else None,
            "verified_data": cbor2.dumps(self.verified_data) if self.verified_data else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def oid(self) -> uuid.UUID:
        """
        Globally unique and static ID of action.
        """
        if self._oid is None and self._from_fbs:
            if self._from_fbs.OidLength():
                _oid = self._from_fbs.OidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
            else:
                self._oid = uuid.UUID(bytes=b"\x00" * 20)
        return self._oid

    @oid.setter
    def oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._oid = value

    @property
    def created(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) of initial creation of this record.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def vtype(self) -> int:
        """
        Verification type.
        """
        if self._vtype is None and self._from_fbs:
            self._vtype = self._from_fbs.Vtype()
        return self._vtype or 0

    @vtype.setter
    def vtype(self, value: int):
        assert value is None or (type(value) == int and value in self.VERIFICATION_TYPE)
        self._vtype = value

    @property
    def vstatus(self) -> int:
        """
        Verification type.
        """
        if self._vstatus is None and self._from_fbs:
            self._vstatus = self._from_fbs.Vstatus()
        return self._vstatus or 0

    @vstatus.setter
    def vstatus(self, value: int):
        assert value is None or (type(value) == int and value in self.VERIFICATION_STATUS)
        self._vstatus = value

    @property
    def vcode(self) -> str:
        """
        Verification code sent.
        """
        if self._vcode is None and self._from_fbs:
            vcode = self._from_fbs.Vcode()
            if vcode:
                self._vcode = vcode.decode("utf8")
        return self._vcode

    @vcode.setter
    def vcode(self, value: str):
        assert value is None or type(value) == str
        self._vcode = value

    @property
    def verified_oid(self) -> uuid.UUID:
        """
        ID of object of verified action.
        """
        if self._verified_oid is None and self._from_fbs:
            if self._from_fbs.VerifiedOidLength():
                _verified_oid = self._from_fbs.VerifiedOidAsBytes()
                self._verified_oid = uuid.UUID(bytes=bytes(_verified_oid))
            else:
                self._verified_oid = uuid.UUID(bytes=b"\x00" * 16)
        return self._verified_oid

    @verified_oid.setter
    def verified_oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._verified_oid = value

    @property
    def verified_data(self) -> dict:
        """
        Action data, serialized in CBOR.
        """
        if self._verified_data is None and self._from_fbs:
            _verified_data = self._from_fbs.VerifiedDataAsBytes()
            if _verified_data:
                self._verified_data = cbor2.loads(_verified_data)
            else:
                self._verified_data = {}
        return self._verified_data

    @verified_data.setter
    def verified_data(self, value: dict):
        assert value is None or type(value) == dict
        self._verified_data = value

    @staticmethod
    def cast(buf):
        return VerifiedAction(_VerifiedActionGen.GetRootAsVerifiedAction(buf, 0))

    def build(self, builder):
        oid = self.oid.bytes if self.oid else None
        if oid:
            oid = builder.CreateString(oid)

        vcode = self.vcode
        if vcode:
            vcode = builder.CreateString(vcode)

        verified_oid = self.verified_oid.bytes if self.verified_oid else None
        if verified_oid:
            verified_oid = builder.CreateString(verified_oid)

        verified_data = self.verified_data
        if verified_data:
            verified_data = builder.CreateString(cbor2.dumps(verified_data))

        VerifiedActionGen.VerifiedActionStart(builder)

        if oid:
            VerifiedActionGen.VerifiedActionAddOid(builder, oid)

        if self.created:
            VerifiedActionGen.VerifiedActionAddCreated(builder, int(self.created))

        if self.vtype:
            VerifiedActionGen.VerifiedActionAddVtype(builder, self.vtype)

        if self.vstatus:
            VerifiedActionGen.VerifiedActionAddVstatus(builder, self.vstatus)

        if vcode:
            VerifiedActionGen.VerifiedActionAddVcode(builder, vcode)

        if verified_oid:
            VerifiedActionGen.VerifiedActionAddVerifiedOid(builder, verified_oid)

        if verified_data:
            VerifiedActionGen.VerifiedActionAddVerifiedData(builder, verified_data)

        final = VerifiedActionGen.VerifiedActionEnd(builder)

        return final


@table("84d805a7-c012-4dfa-9b95-e34760767f82", build=VerifiedAction.build, cast=VerifiedAction.cast)
class VerifiedActions(MapUuidFlatBuffers):
    """
    Database table for verification/verified actions, eg on-boarding new XBR members.
    """

    @staticmethod
    def parse(data):
        """

        :param data:
        :return:
        """
        oid = None
        if "oid" in data:
            assert type(data["oid"] == bytes and len(data["oid"]) == 16)
            oid = uuid.UUID(bytes=data["oid"])

        created = None
        if "created" in data:
            assert type(data["created"] == int)
            created = np.datetime64(data["created"], "ns")

        vtype = None
        if "vtype" in data:
            assert type(data["vtype"] == int)
            vtype = data["vtype"]

        vstatus = None
        if "vstatus" in data:
            assert type(data["vstatus"] == int)
            vstatus = data["vstatus"]

        vcode = None
        if "vcode" in data:
            assert type(data["vcode"] == str)
            vcode = data["vcode"]

        verified_oid = None
        if "verified_oid" in data:
            assert type(data["verified_oid"] == bytes and len(data["verified_oid"]) == 16)
            verified_oid = uuid.UUID(bytes=data["verified_oid"])

        verified_data = None
        if "verified_data" in data:
            assert type(data["verified_data"]) == bytes
            verified_data = cbor2.loads(data["verified_data"])

        obj = VerifiedAction()
        obj.oid = oid
        obj.created = created
        obj.vtype = vtype
        obj.vstatus = vstatus
        obj.vcode = vcode
        obj.verified_oid = verified_oid
        obj.verified_data = verified_data

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log


class MasterNodeUsageStatus(object):
    # Unset state.
    NONE = 0
    # Online and fully operational.
    RECEIVED = 1
    # Online, but some error occurred and some features or entities might be unavailable.
    SUCCESS = 2
    FAILED = 3



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR Markets.
class Market(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Market()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMarket(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Market
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The unique ID of the market.
    # Market
    def Market(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def MarketAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def MarketLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def MarketIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Database transaction time (epoch time in ns) of insert or last update.
    # Market
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number (on the blockchain) when the actor (originally) joined the market.
    # Market
    def Created(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def CreatedAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def CreatedLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def CreatedIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Global market sequence number.
    # Market
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Market owner.
    # Market
    def Owner(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def OwnerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def OwnerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def OwnerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Market coin used as means of payment.
    # Market
    def Coin(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def CoinAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def CoinLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def CoinIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # The XBR market terms set by the market owner. IPFS Multihash pointing to a ZIP archive file with market documents.
    # Market
    def Terms(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
    # Market
    def Meta(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The address of the XBR market maker that will run this market. The delegate of the market owner.
    # Market
    def Maker(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def MakerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def MakerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def MakerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # The amount of XBR tokens a XBR provider joining the market must deposit.
    # Market
    def ProviderSecurity(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def ProviderSecurityAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def ProviderSecurityLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def ProviderSecurityIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # The amount of XBR tokens a XBR consumer joining the market must deposit.
    # Market
    def ConsumerSecurity(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def ConsumerSecurityAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def ConsumerSecurityLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def ConsumerSecurityIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # The fee taken by the market (beneficiary is the market owner). The fee is a percentage of the revenue of the XBR Provider that receives XBR Token paid for transactions. The fee must be between 0% (inclusive) and 99% (inclusive), and is expressed as a fraction of the total supply of XBR tokens.
    # Market
    def MarketFee(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def MarketFeeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def MarketFeeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def MarketFeeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Market
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # When signed off-chain and submitted via ``XBRMarket.createMarketFor``.
    # Market
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Market
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Market
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Market
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0


def MarketStart(builder):
    builder.StartObject(14)


def Start(builder):
    MarketStart(builder)


def MarketAddMarket(builder, market):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(market), 0)


def AddMarket(builder, market):
    MarketAddMarket(builder, market)


def MarketStartMarketVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketVector(builder, numElems: int) -> int:
    return MarketStartMarketVector(builder, numElems)


def MarketAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(1, timestamp, 0)


def AddTimestamp(builder, timestamp):
    MarketAddTimestamp(builder, timestamp)


def MarketAddCreated(builder, created):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(created), 0)


def AddCreated(builder, created):
    MarketAddCreated(builder, created)


def MarketStartCreatedVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCreatedVector(builder, numElems: int) -> int:
    return MarketStartCreatedVector(builder, numElems)


def MarketAddSeq(builder, seq):
    builder.PrependUint32Slot(3, seq, 0)


def AddSeq(builder, seq):
    MarketAddSeq(builder, seq)


def MarketAddOwner(builder, owner):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    MarketAddOwner(builder, owner)


def MarketStartOwnerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOwnerVector(builder, numElems: int) -> int:
    return MarketStartOwnerVector(builder, numElems)


def MarketAddCoin(builder, coin):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(coin), 0)


def AddCoin(builder, coin):
    MarketAddCoin(builder, coin)


def MarketStartCoinVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCoinVector(builder, numElems: int) -> int:
    return MarketStartCoinVector(builder, numElems)


def MarketAddTerms(builder, terms):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(terms), 0)


def AddTerms(builder, terms):
    MarketAddTerms(builder, terms)


def MarketAddMeta(builder, meta):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(meta), 0)


def AddMeta(builder, meta):
    MarketAddMeta(builder, meta)


def MarketAddMaker(builder, maker):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(maker), 0)


def AddMaker(builder, maker):
    MarketAddMaker(builder, maker)


def MarketStartMakerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMakerVector(builder, numElems: int) -> int:
    return MarketStartMakerVector(builder, numElems)


def MarketAddProviderSecurity(builder, providerSecurity):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(providerSecurity), 0)


def AddProviderSecurity(builder, providerSecurity):
    MarketAddProviderSecurity(builder, providerSecurity)


def MarketStartProviderSecurityVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartProviderSecurityVector(builder, numElems: int) -> int:
    return MarketStartProviderSecurityVector(builder, numElems)


def MarketAddConsumerSecurity(builder, consumerSecurity):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(consumerSecurity), 0)


def AddConsumerSecurity(builder, consumerSecurity):
    MarketAddConsumerSecurity(builder, consumerSecurity)


def MarketStartConsumerSecurityVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartConsumerSecurityVector(builder, numElems: int) -> int:
    return MarketStartConsumerSecurityVector(builder, numElems)


def MarketAddMarketFee(builder, marketFee):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(marketFee), 0)


def AddMarketFee(builder, marketFee):
    MarketAddMarketFee(builder, marketFee)


def MarketStartMarketFeeVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketFeeVector(builder, numElems: int) -> int:
    return MarketStartMarketFeeVector(builder, numElems)


def MarketAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    MarketAddTid(builder, tid)


def MarketStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return MarketStartTidVector(builder, numElems)


def MarketAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    MarketAddSignature(builder, signature)


def MarketStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return MarketStartSignatureVector(builder, numElems)


def MarketEnd(builder):
    return builder.EndObject()


def End(builder):
    return MarketEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log


class IpVersion(object):
    # Unset IP version.
    NONE = 0
    # IP version 4.
    IPV4 = 4
    # IP version 6.
    IPV6 = 6



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import Optional
from uuid import UUID

from cfxdb.mrealm.cluster_node_membership import ClusterNodeMembership


class WebClusterNodeMembership(ClusterNodeMembership):
    """
    Information about memberships of nodes in web clusters.
    """

    def __init__(
        self,
        cluster_oid: Optional[UUID] = None,
        node_oid: Optional[UUID] = None,
        parallel: Optional[int] = None,
        standby: Optional[bool] = None,
        _unknown=None,
    ):
        """

        :param cluster_oid: Object ID of the cluster the node is member in.

        :param node_oid: Object ID of the node that is member in the cluster.

        :param parallel: Default parallel degree per node in this cluster.

        :param standby: Flag indicating whether this is a (currently inactive) standby node.
        """
        ClusterNodeMembership.__init__(self, cluster_oid=cluster_oid, node_oid=node_oid, _unknown=_unknown)
        self.parallel = parallel
        self.standby = standby

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ClusterNodeMembership.__eq__(self, other):
            return False
        if other.parallel != self.parallel:
            return False
        if other.standby != self.standby:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ClusterNodeMembership.marshal(self)
        obj.update(
            {
                "parallel": self.parallel,
                "standby": self.standby,
            }
        )
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        obj = ClusterNodeMembership.parse(data)
        data = obj._unknown or {}

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["parallel", "standby"]:
                _unknown[k] = data[k]

        parallel = None
        if "parallel" in data and data["parallel"]:
            assert type(data["parallel"]) == int
            parallel = data["parallel"]

        standby = None
        if "standby" in data and data["standby"]:
            assert data["standby"] is None or type(data["standby"]) == bool
            standby = data["standby"]

        obj = WebClusterNodeMembership(
            cluster_oid=obj.cluster_oid, node_oid=obj.node_oid, parallel=parallel, standby=standby, _unknown=_unknown
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork


class VerificationType(object):
    # Unset
    NONE = 0
    # Member onboarded, using verification via member primary email.
    MEMBER_ONBOARD_EMAIL = 1
    # Member logged in, using verification via member primary email.
    MEMBER_LOGIN_EMAIL = 2
    # Member created a market, using verification via member primary email.
    MEMBER_CREATEMARKET_EMAIL = 3
    # Member joined a market, using verification via member primary email.
    MEMBER_JOINMARKET_EMAIL = 4



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from autobahn import util
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.cookiestore import Cookie


def fill_cookie(cookie):
    cookie.oid = uuid.uuid4()
    cookie.created = np.datetime64(time_ns(), "ns")
    cookie.max_age = random.randint(1, 10**10)
    cookie.name = random.choice(["cbtid1", "cbtid2", "cbtid3"])
    cookie.value = util.newid(24)
    cookie.authenticated = np.datetime64(time_ns(), "ns")
    cookie.authenticated_on_node = uuid.uuid4()
    cookie.authenticated_on_worker = random.choice(["worker1", "worker2", "worker3"])
    cookie.authenticated_transport_info = {"xoo": "yar", "zaz": [9, 8, 7]}
    cookie.authenticated_session = util.id()
    cookie.authenticated_joined_at = np.datetime64(time_ns(), "ns")
    cookie.authenticated_authmethod = random.choice(["meth1", "meth2", "meth3"])
    cookie.authid = util.generate_token(4, 3)
    cookie.authrole = random.choice(["role1", "role2", "role3"])
    cookie.authmethod = random.choice(["method1", "method2", "method3"])
    cookie.authrealm = random.choice(["realm1", "realm2", "realm3"])
    cookie.authextra = {"foo": "bar", "baz": [1, 2, 3]}


def fill_cookie_empty(cookie):
    cookie.oid = None
    cookie.created = None
    cookie.max_age = None
    cookie.name = None
    cookie.value = None
    cookie.authenticated = None
    cookie.authenticated_on_node = None
    cookie.authenticated_on_worker = None
    cookie.authenticated_transport_info = None
    cookie.authenticated_session = None
    cookie.authenticated_joined_at = None
    cookie.authenticated_authmethod = None
    cookie.authid = None
    cookie.authrole = None
    cookie.authmethod = None
    cookie.authrealm = None
    cookie.authextra = None


@pytest.fixture(scope="function")
def cookie():
    _cookie = Cookie()
    fill_cookie(_cookie)
    return _cookie


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_cookie_roundtrip(cookie, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = cookie.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 360

    # create python object from bytes (flatbuffes)
    _cookie = Cookie.cast(data)

    assert _cookie.oid == cookie.oid
    assert _cookie.created == cookie.created
    assert _cookie.max_age == cookie.max_age
    assert _cookie.name == cookie.name
    assert _cookie.value == cookie.value
    assert _cookie.authenticated == cookie.authenticated
    assert _cookie.authenticated_on_node == cookie.authenticated_on_node
    assert _cookie.authenticated_on_worker == cookie.authenticated_on_worker
    assert _cookie.authenticated_transport_info == cookie.authenticated_transport_info
    assert _cookie.authenticated_session == cookie.authenticated_session
    assert _cookie.authenticated_joined_at == cookie.authenticated_joined_at
    assert _cookie.authenticated_authmethod == cookie.authenticated_authmethod
    assert _cookie.authid == cookie.authid
    assert _cookie.authrole == cookie.authrole
    assert _cookie.authmethod == cookie.authmethod
    assert _cookie.authrealm == cookie.authrealm
    assert _cookie.authextra == cookie.authextra


def test_cookie_empty(builder):
    cookie = Cookie()
    fill_cookie_empty(cookie)

    # serialize to bytes (flatbuffers) from python object
    obj = cookie.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 12

    # create python object from bytes (flatbuffes)
    _cookie = Cookie.cast(data)

    unix_zero = np.datetime64(0, "ns")

    assert _cookie.oid is None
    assert _cookie.created == unix_zero
    assert _cookie.max_age == 0
    assert _cookie.name is None
    assert _cookie.value is None
    assert _cookie.authenticated == unix_zero
    assert _cookie.authenticated_on_node is None
    assert _cookie.authenticated_on_worker is None
    assert _cookie.authenticated_transport_info == {}
    assert _cookie.authenticated_session == 0
    assert _cookie.authenticated_joined_at == unix_zero
    assert _cookie.authenticated_authmethod is None
    assert _cookie.authid is None
    assert _cookie.authrole is None
    assert _cookie.authmethod is None
    assert _cookie.authrealm is None
    assert _cookie.authextra == {}


def test_cookie_roundtrip_perf(cookie, builder):
    obj = cookie.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _cookie = Cookie.cast(data)

        assert _cookie.oid == cookie.oid
        assert _cookie.created == cookie.created
        assert _cookie.max_age == cookie.max_age
        assert _cookie.name == cookie.name
        assert _cookie.value == cookie.value
        assert _cookie.authenticated == cookie.authenticated
        assert _cookie.authenticated_on_node == cookie.authenticated_on_node
        assert _cookie.authenticated_on_worker == cookie.authenticated_on_worker
        assert _cookie.authenticated_transport_info == cookie.authenticated_transport_info
        assert _cookie.authenticated_session == cookie.authenticated_session
        assert _cookie.authenticated_joined_at == cookie.authenticated_joined_at
        assert _cookie.authenticated_authmethod == cookie.authenticated_authmethod
        assert _cookie.authid == cookie.authid
        assert _cookie.authrole == cookie.authrole
        assert _cookie.authmethod == cookie.authmethod
        assert _cookie.authrealm == cookie.authrealm
        assert _cookie.authextra == cookie.authextra

        scratch["value"] += _cookie.authenticated_session

    N = 7
    M = 20000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################
import pprint
import uuid

import numpy as np
from zlmdb import MapBytes16TimestampUuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Api as ApiGen


class _ApiGen(ApiGen.Api):
    @classmethod
    def GetRootAsApi(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ApiGen()
        x.Init(buf, n + offset)
        return x

    def OidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CatalogOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PublishedAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Api(object):
    """
    ``XBRNetwork.Api`` database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (uuid);
        self._oid = None

        # [uint8] (uuid);
        self._catalog_oid = None

        # uint64 (timestamp)
        self._timestamp = None

        # [uint8] (uint256)
        self._published = None

        # string (multihash)
        self._schema = None

        # string (multihash)
        self._meta = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

    def marshal(self) -> dict:
        obj = {
            "oid": self.oid.bytes if self.oid else None,
            "catalog_oid": self.catalog_oid.bytes if self.catalog_oid else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "published": pack_uint256(self.published) if self.published else None,
            "schema": self.schema,
            "meta": self.meta,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def oid(self) -> uuid.UUID:
        """
        The unique ID of the api.
        """
        if self._oid is None and self._from_fbs:
            if self._from_fbs.OidLength():
                _oid = self._from_fbs.OidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
        return self._oid

    @oid.setter
    def oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._oid = value

    @property
    def catalog_oid(self) -> uuid.UUID:
        """
        The unique ID of the catalog.
        """
        if self._catalog_oid is None and self._from_fbs:
            if self._from_fbs.CatalogOidLength():
                _oid = self._from_fbs.CatalogOidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
        return self._oid

    @catalog_oid.setter
    def catalog_oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._catalog_oid = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def published(self) -> int:
        """
        Global market sequence number.
        """
        if self._published is None and self._from_fbs:
            if self._from_fbs.PublishedLength():
                _published = self._from_fbs.PublishedAsBytes()
                self._published = unpack_uint256(bytes(_published))
            else:
                self._published = 0
        return self._published

    @published.setter
    def published(self, value: int):
        assert value is None or type(value) == int
        self._published = value

    @property
    def schema(self) -> str:
        """
        The XBR market terms set by the market owner. IPFS Multihash pointing to a ZIP archive file with market documents.
        """
        if self._schema is None and self._from_fbs:
            schema = self._from_fbs.Schema()
            if schema:
                self._schema = schema.decode("utf8")
        return self._schema

    @schema.setter
    def schema(self, value: str):
        assert value is None or type(value) == str
        self._schema = value

    @property
    def meta(self) -> str:
        """
        The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
        """
        if self._meta is None and self._from_fbs:
            meta = self._from_fbs.Meta()
            if meta:
                self._meta = meta.decode("utf8")
        return self._meta

    @meta.setter
    def meta(self, value):
        assert value is None or type(value) == str
        self._meta = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRMarket.createMarketFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return Api(_ApiGen.GetRootAsApi(buf, 0))

    def build(self, builder):
        oid = self.oid.bytes if self.oid else None
        if oid:
            oid = builder.CreateString(oid)

        catalog_oid = self.catalog_oid.bytes if self.catalog_oid else None
        if catalog_oid:
            catalog_oid = builder.CreateString(catalog_oid)

        published = self.published
        if published:
            published = builder.CreateString(pack_uint256(published))

        schema = self.schema
        if schema:
            schema = builder.CreateString(schema)

        meta = self.meta
        if meta:
            meta = builder.CreateString(meta)

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        ApiGen.ApiStart(builder)

        if oid:
            ApiGen.ApiAddOid(builder, oid)

        if catalog_oid:
            ApiGen.ApiAddCatalogOid(builder, catalog_oid)

        if self.timestamp:
            ApiGen.ApiAddTimestamp(builder, int(self.timestamp))

        if published:
            ApiGen.ApiAddPublished(builder, published)

        if schema:
            ApiGen.ApiAddSchema(builder, schema)

        if meta:
            ApiGen.ApiAddMeta(builder, meta)

        if tid:
            ApiGen.ApiAddTid(builder, tid)

        if signature:
            ApiGen.ApiAddSignature(builder, signature)

        final = ApiGen.ApiEnd(builder)

        return final


@table("9f87bad9-695e-439d-86ad-9e23695d3f67", build=Api.build, cast=Api.cast)
class Apis(MapUuidFlatBuffers):
    """
    Apis table, mapping from ``oid|UUID`` to :class:`cfxdb.xbr.Api`
    """


@table("57ae5d83-57b0-4214-8481-a1853f7faf87")
class IndexApiByCatalog(MapBytes16TimestampUuid):
    """
    Api-by-FbsRepository, index with ``(catalog_oid|bytes[16], created|int) -> api_oid|UUID`` mapping.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR catalogs.
class Catalog(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Catalog()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCatalog(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Catalog
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The unique ID of the catalog.
    # Catalog
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Catalog
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Catalog
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Catalog
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Database transaction time (epoch time in ns).
    # Catalog
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number when the catalog was created.
    # Catalog
    def Created(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Catalog
    def CreatedAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Catalog
    def CreatedLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Catalog
    def CreatedIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # FbsRepository sequence number.
    # Catalog
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # FbsRepository owner (aka "catalog publisher").
    # Catalog
    def Owner(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Catalog
    def OwnerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Catalog
    def OwnerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Catalog
    def OwnerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # FbsRepository terms (IPFS Multihash).
    # Catalog
    def Terms(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FbsRepository metadata (IPFS Multihash).
    # Catalog
    def Meta(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Catalog
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Catalog
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Catalog
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Catalog
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # When signed off-chain and submitted via ``XBRCatalog.createCatalogFor``.
    # Catalog
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Catalog
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Catalog
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Catalog
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0


def CatalogStart(builder):
    builder.StartObject(9)


def Start(builder):
    CatalogStart(builder)


def CatalogAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    CatalogAddOid(builder, oid)


def CatalogStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return CatalogStartOidVector(builder, numElems)


def CatalogAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(1, timestamp, 0)


def AddTimestamp(builder, timestamp):
    CatalogAddTimestamp(builder, timestamp)


def CatalogAddCreated(builder, created):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(created), 0)


def AddCreated(builder, created):
    CatalogAddCreated(builder, created)


def CatalogStartCreatedVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCreatedVector(builder, numElems: int) -> int:
    return CatalogStartCreatedVector(builder, numElems)


def CatalogAddSeq(builder, seq):
    builder.PrependUint32Slot(3, seq, 0)


def AddSeq(builder, seq):
    CatalogAddSeq(builder, seq)


def CatalogAddOwner(builder, owner):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    CatalogAddOwner(builder, owner)


def CatalogStartOwnerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOwnerVector(builder, numElems: int) -> int:
    return CatalogStartOwnerVector(builder, numElems)


def CatalogAddTerms(builder, terms):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(terms), 0)


def AddTerms(builder, terms):
    CatalogAddTerms(builder, terms)


def CatalogAddMeta(builder, meta):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(meta), 0)


def AddMeta(builder, meta):
    CatalogAddMeta(builder, meta)


def CatalogAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    CatalogAddTid(builder, tid)


def CatalogStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return CatalogStartTidVector(builder, numElems)


def CatalogAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    CatalogAddSignature(builder, signature)


def CatalogStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return CatalogStartSignatureVector(builder, numElems)


def CatalogEnd(builder):
    return builder.EndObject()


def End(builder):
    return CatalogEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from datetime import datetime
from typing import List, Optional
from uuid import UUID

from cfxdb.common import ConfigurationElement


class ManagementRealm(ConfigurationElement):
    """
    Management realm database configuration object.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        created: Optional[datetime] = None,
        owner: Optional[UUID] = None,
        cf_router_worker: Optional[str] = None,
        cf_container_worker: Optional[str] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of management realm

        :param label: Optional user label of management realm

        :param description: Optional user description of management realm

        :param tags: Optional list of user tags on management realm

        :param name: Name of management realm

        :param created: Timestamp when the management realm was created

        :param owner: Owning user (object ID)

        :param cf_router_worker: *INTERNAL USE* CFC hosting router worker for this management realm

        :param cf_container_worker: *INTERNAL USE* CFC hosting container worker for this management realm
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.name = name
        self.created = created
        self.owner = owner
        self.cf_router_worker = cf_router_worker
        self.cf_container_worker = cf_container_worker

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.name != self.name:
            return False
        if other.created != self.created:
            return False
        if other.owner != self.owner:
            return False
        if other.cf_router_worker != self.cf_router_worker:
            return False
        if other.cf_container_worker != self.cf_container_worker:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other management realm to copy data from.
        :type other: instance of :class:`ManagementRealm`
        :return:
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.name and other.name) or overwrite:
            self.name = other.name
        if (not self.created and other.created) or overwrite:
            self.created = other.created
        if (not self.owner and other.owner) or overwrite:
            self.owner = other.owner
        if (not self.cf_router_worker and other.cf_router_worker) or overwrite:
            self.cf_router_worker = other.cf_router_worker
        if (not self.cf_container_worker and other.cf_container_worker) or overwrite:
            self.cf_container_worker = other.cf_container_worker

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "oid": str(self.oid),
                "name": self.name,
                "created": int(self.created.timestamp() * 1000000) if self.created else None,
                "owner": str(self.owner),
                "cf_router_worker": self.cf_router_worker,
                "cf_container_worker": self.cf_container_worker,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "name", "rtype", "owner", "created", "cf_router_worker", "cf_container_worker"]:
                _unknown[k] = data[k]

        name = data.get("name", None)
        assert name is None or type(name) == str

        owner = data.get("owner", None)
        assert owner is None or type(owner) == str
        if owner:
            owner = UUID(owner)

        created = data.get("created", None)
        assert created is None or type(created) == float or type(created) == int
        if created:
            created = datetime.utcfromtimestamp(float(created) / 1000000.0)

        cf_router_worker = data.get("cf_router_worker", None)
        assert cf_router_worker is None or type(cf_router_worker) == str

        cf_container_worker = data.get("cf_container_worker", None)
        assert cf_container_worker is None or type(cf_container_worker) == str

        obj = ManagementRealm(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=name,
            owner=owner,
            created=created,
            cf_router_worker=cf_router_worker,
            cf_container_worker=cf_container_worker,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import pytest
import txaio
import zlmdb
from zlmdb import flatbuffers

txaio.use_twisted()  # noqa

from txaio import time_ns

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr.api import Api

zlmdb.TABLES_BY_UUID = {}


def fill_api(api):
    api.oid = uuid.uuid4()
    api.catalog_oid = uuid.uuid4()
    api.timestamp = np.datetime64(time_ns(), "ns")
    api.published = random.randint(0, 2**256 - 1)
    api.schema = _gen_ipfs_hash()
    api.meta = _gen_ipfs_hash()
    api.tid = os.urandom(32)
    api.signature = os.urandom(65)


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def api():
    _api = Api()
    fill_api(_api)
    return _api


def test_api_roundtrip(api: Api, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = api.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 368

    # create python object from bytes (flatbuffes)
    _api = Api.cast(data)

    assert _api.oid == api.oid
    assert _api.catalog_oid == api.catalog_oid
    assert _api.timestamp == api.timestamp
    assert _api.published == api.published
    assert _api.schema == api.schema
    assert _api.meta == api.meta
    assert _api.tid == api.tid
    assert _api.signature == api.signature


def test_api_roundtrip_perf(api, builder):
    obj = api.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _api = Api.cast(data)
        if True:
            assert _api.oid == api.oid
            assert _api.catalog_oid == api.catalog_oid
            assert _api.timestamp == api.timestamp
            assert _api.published == api.published
            assert _api.schema == api.schema
            assert _api.meta == api.meta
            assert _api.tid == api.tid
            assert _api.signature == api.signature

            scratch["value"] += _api.published

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])
